package markov;

import java.util.AbstractMap;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.Function;

public class Builder {

    private final int prefix_length;
    private final Data data;

    public Builder(int prefix_length, Data data) {
        super();
        this.prefix_length = prefix_length;
        this.data = data;
    }

    public Sentence random() {
        return produce(this::nextRandom);
    }

    public Sentence average() {
        return produce(p -> data.fetch(p).average());
    }

    private Sentence produce(
            Function<Prefix, Decission> producer) {
        Sentence sentence = new Sentence(data);
        Prefix p = initPrefix();
        Decission d = producer.apply(p);
        for (int i = 0; i < 1000; i++) {
            if (Token.END.equals(d.getToken()))
                break;
            d = producer.apply(p);
            sentence.add(d);
            p = p.slide(d.getToken(), prefix_length);
        }
        return sentence;
    }

    private Prefix initPrefix() {
        return new Prefix(new Token[0]);
    }

    private Decission nextRandom(Prefix prefix) {
        Lookup l = data.fetch(prefix);
        if (l == null) {
            throw new NullPointerException(String.format("could not find a lookup for %s", prefix));
        }
        Decission decission = l.forRandom(nextRandomNumber());
        return decission;
    }

    private double nextRandomNumber() {
        return ThreadLocalRandom.current().nextDouble();
    }
}
