package markov;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;

// FIXME rename
public class Prefix implements Serializable, Comparable<Prefix> {
    private final Token[] tokens;

    public Prefix(Token[] tokens) {
        this.tokens = tokens;
        this.hashCode = generateHashCode();
    }

    public Prefix(List<Token> tokens) {
        this(tokens.toArray(new Token[tokens.size()]));
    }

    public Prefix slide(Token newToken) {
        return slide(newToken, this.tokens.length);
    }

    public Prefix slide(Token newToken, int maxLength) {
        int newLength = Math.min(tokens.length + 1, maxLength);
        Token[] newTokens = new Token[newLength];
        System.arraycopy(tokens, Math.max(0, tokens.length - newLength + 1), newTokens, 0,
                         newLength - 1);
        newTokens[newLength - 1] = newToken;
        return new Prefix(newTokens);
    }

    @Override
    public String toString() {
        return "Prefix{" + "tokens=" + Arrays.toString(tokens) + '}';
    }


    private final int hashCode;

    private int generateHashCode() {
        return Arrays.hashCode(tokens);
    }

    @Override
    public int hashCode() {
        return hashCode;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Prefix prefix = (Prefix) o;
        return Arrays.equals(tokens, prefix.tokens);
    }


    @Override
    public int compareTo(Prefix o) {
        if (this.tokens.length != o.tokens.length) {
            return Integer.compare(o.tokens.length, this.tokens.length);
        }
        for (int i = 0; i < this.tokens.length; i++) {
            int c = this.tokens[i].compareTo(o.tokens[i]);
            if (c != 0) {
                return c;
            }
        }
        return 0;
    }
}
