package markov;

import java.nio.ByteBuffer;
import java.util.Base64;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class ShortenerSimpleImpl implements Shortener {

    private final Data data;

    public ShortenerSimpleImpl(Data data) {
        this.data = data;
    }

    @Override
    public String getId(Sentence sentence) {
        return fromInts(sentence.asStream()
                .flatMap(d -> {
                    if (d.getLookup().getDistinctTokens() == 1) {
                        return Stream.empty();
                    } else
                        return Stream.of(d.getId());
                }));

    }

    String fromInts(Stream<Integer> ints) {
        List<Integer> b1 = ints.collect(Collectors.toList());
        ByteBuffer bb = ByteBuffer.allocate(b1.size()*4);
        for(int b : b1) {
            bb.putInt(b);
        }
        bb.rewind();
        return Base64.getEncoder().encodeToString(bb.array());
    }

    Stream<Integer> toInts(String hash) {
        return new String(Base64.getDecoder().decode(hash.getBytes())).codePoints().boxed();
    }

    @Override
    public Sentence getSentence(String hash) {
        Integer[] ids = toInts(hash).toArray(Integer[]::new);

        Sentence sentence = new Sentence(getData());
        Prefix prefix = new Prefix(new Token[0]);
        int i = 0;
        while (i < ids.length) {
            Lookup lookup = getData().fetch(prefix);
            Decission d = lookup.getDistinctTokens() == 1 ? lookup.average() : lookup.forId(ids[i++]);
            sentence.add(d);
            prefix = prefix.slide(d.getToken(), getData().getPrefixLength());
        }
        return sentence;
    }

    protected Data getData() {
        return data;
    }
}
