package markov;

import java.io.Serializable;

public class Token implements Serializable, Comparable<Token> {

    public static final Token START = new SpecialToken("START");
    public static final Token END = new SpecialToken("END");
    protected static final Token EMPTY = new SpecialToken("EMPTY");

    public Token(String content) {
        this(content, Glyph.Type.word);
    }

    public Token(String content, Glyph.Type type) {
        super();
        this.content = content;
        this.hashCode = generateHashCode();
    }

    private String content;
    private Glyph.Type type;

    @Override
    public String toString() {
        if (this.equals(END)) {
            return "Token <END>";
        }
        if (this.equals(START)) {
            return "TOKEN <Start>";
        }
        if (this.equals(EMPTY)) {
            return "TOKEN <empty>";
        }
        return "T=[" + content + "]";
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Token token = (Token) o;

        return (content != null ? content.equals(token.content) : token.content == null) &&
               type == token.type;
    }

    private final int hashCode;

    private int generateHashCode() {
        int result = content != null ? content.hashCode() : 0;
        result = 31 * result + (type != null ? type.hashCode() : 0);
        return result;
    }

    @Override
    public int hashCode() {
        return hashCode;
    }

    public String render(String prefix) {
        return prefix + content;
    }

    @Override
    public int compareTo(Token o) {
        return o.content.compareTo(this.content);
    }

    private static class SpecialToken extends Token {

        public SpecialToken(String s) {
            super(s, Glyph.Type.control);
        }

        @Override
        public String render(String prefix) {
            return this.toString();
        }

    }

}
