package markov.huffman;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class BitConverter {


    public byte[] toBytes(List<Boolean> list) {
        int listLength = list.size();
        int lengthInBit = listLength + 3;
        int arrayLength = Math.max(1, (int) Math.ceil(lengthInBit / 8d));
        byte[] bytes = new byte[arrayLength];
        int lastByteNotEmpty = lengthInBit % 8;
        // temp variable to not constantly write to heap
        byte tmpByte = ((byte) lastByteNotEmpty);
        int i = 0;
        // set 3 bits for the lastByteNotEmpty number
        for (int j = 3; j < 8 && i < listLength; j++, i++) {
            tmpByte = setBit(list, tmpByte, i, j);
        }
        bytes[0] = tmpByte;
        // set rest of the bits
        for (int j = 1; j < arrayLength; j++) {
            tmpByte = 0;
            for (int l = 0; l < 8 && i < listLength; l++, i++) {
                tmpByte = setBit(list, tmpByte, i, l);
            }
            bytes[j] = tmpByte;
        }
        return bytes;
    }

    private byte setBit(List<Boolean> list, byte tmpByte, int i, int j) {
        tmpByte |= (byte) (list.get(i) ? 1 : 0) << j;
        return tmpByte;
    }

    public List<Boolean> toBits(byte[] array) {
        int arrayLength = array.length;
        if (arrayLength == 0) {
            return Collections.emptyList();
        }
        // determine lastByteNotEmpty bits
        byte tmpByte = array[0];
        int lastByteNotEmpty = tmpByte & 0b00000111;
        int listSize = 8 * arrayLength + (lastByteNotEmpty == 0 ? -3 : lastByteNotEmpty - 11);
        if (arrayLength == 1 && lastByteNotEmpty < 4 && lastByteNotEmpty != 0) {
            throw new IllegalArgumentException("corrupted data");
        }
        return toListPrimitive(listSize, array);
    }

    private static List<Boolean> toListPrimitive(int listSize, byte[] array) {
        List<Boolean> list = new ArrayList<>(listSize);
        for (int i = 3; i < listSize + 3; i++) {
            list.add((array[i / 8] & 1 << (i % 8)) >> i % 8 == 1);
        }
        return list;
    }

}
