package markov.stuff;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.stream.Stream;

public class SimpleCountMap<T> extends HashMap<T, SimpleCountMap.MutableInt>
        implements CountMap<T> {
    public SimpleCountMap(int initialCapacity, float loadFactor) {
        super(initialCapacity, loadFactor);
    }


    public SimpleCountMap(int initialCapacity) {
        super(initialCapacity);
    }

    public SimpleCountMap() {
    }

    public SimpleCountMap(Map<? extends T, ? extends MutableInt> m) {
        super(m);
    }

    @Override
    public void count(T t) {
        MutableInt i = get(t);
        if (i == null) {
            put(t, new MutableInt(1));
        } else {
            i.increment();
        }
    }

    @Override
    public Stream<Map.Entry<T, MutableInt>> result() {
        return entrySet().stream().sorted((e1, e2) -> Integer.compare(e2.getValue().get(),
                e1.getValue().get()));
    }

    @Override
    public HashMap<T, MutableInt> asMap() {
        return this;
    }

    public static class MutableInt implements Serializable {
        private int value = 1;

        public MutableInt(int value) {
            this.value = value;
        }

        public void increment() {
            ++value;
        }

        public int get() {
            return value;
        }

        public static class FrequencySupport
                implements directory.passive.huffman.FrequencySupport<MutableInt> {

            @Override
            public MutableInt add(
                    MutableInt mutableInt, MutableInt x1) {
                return new MutableInt(mutableInt.get() + x1.get());
            }

            @Override
            public int compare(MutableInt mutableInt, MutableInt x1) {
                return Integer.compare(mutableInt.get(), x1.get());
            }

            @Override
            public boolean isGreaterZero(MutableInt mutableInt) {
                return mutableInt.get() > 0;
            }
        }

    }
}