package markov.stuff;

import markov.Data;
import markov.Parser;
import markov.Token;
import markov.Tokenizer;

import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.internet.MimeMessage;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public class Utils {
    public static byte[] toByteArray(int value) {
        return ByteBuffer.allocate(4).putInt(value).array();
    }

    public static int fromByteArray(byte[] bytes) {
        return ByteBuffer.wrap(bytes).getInt();
    }

    public static boolean[] fromBitList(List<Boolean> list) {
        boolean[] ret = new boolean[list.size()];
        int i = 0;
        for (boolean b : list) {
            ret[i++] = b;
        }
        return ret;
    }

    public static Data parse(Stream<String> input, int prefixLength) {
        Collection<Integer> collectionDummy = new ArrayList<>();
        collectionDummy.add(prefixLength);
        return new Parser().parse(new Tokenizer().tokenize(input), collectionDummy).values()
                .stream().findFirst().get();
    }

    public static File getFile(String location) {
        File file;
        if (location.startsWith("classpath:")) {
            URL url = Utils.class.getClassLoader().getResource(location.replace("classpath:", ""));
            try {
                file = new File(url.toURI());
            } catch (URISyntaxException e) {
                throw new IllegalArgumentException(e);
            }
        } else {
            file = new File(location);
        }
        if (!file.exists() || !file.canRead() || !file.canWrite()) {
            throw new IllegalArgumentException(String.format("could not not use file: %s", file.getName()));
        }
        return file;
    }

    public static void maybePrintPercentages(final int i, final int max) {
        maybePrintPercentages("", i, max);
    }

    public static void maybePrintPercentages(String prefix, final int i, final int max) {
        if (max > 100_000) {
            if (i % (max / 100) == 0) {
                System.out.println(prefix + i / (max / 100) + "%");
                return;
            }
        }
        if (max < 10) {
            System.out.println(prefix + " " + i + "/" + max);
            return;
        }
        if (max < 100) {
            System.out.println(prefix + (int) ((double) i / max * 100) + "%");
            return;
        }
        if (i % (max / 10) == 0) {
            System.out.println(prefix + i / (max / 100) + "%");
        }
    }

    public static Map<Integer, Data> deserializeDataMap() throws IOException,
            ClassNotFoundException {
        System.out.println("reading map");
        Map<Integer, Data> dataMap = (Map<Integer, Data>) new ObjectInputStream(
                Files.newInputStream(new File(
                        "C:\\Users\\admin\\git\\markovdisplay\\target\\classes\\data").toPath()))
                .readObject();
        System.out.println("read map");
        return dataMap;
    }

    public static Map<Integer, Data> createDataMap(int maxPrefix) {
        return createDataMap(maxPrefix, new File("C:\\Users\\admin\\Desktop\\emails"));
    }

    public static Map<Integer, Data> createDataMap(int maxPrefix, File parent) {
        System.out.println("generating data map...");
        Stream<String> mails = Arrays.stream(parent.listFiles()).limit(500).map(File::toPath).map
                (path -> {
            try {

                MimeMessage m = new MimeMessage(
                        Session.getDefaultInstance(new Properties()), Files.newInputStream(path));
                return m;
            } catch (IOException | MessagingException e) {
                throw new RuntimeException(e);
            }
        }).filter(mimeMessage -> {
            try {
                return mimeMessage.getContentType().contains("text/plain");
            } catch (MessagingException e) {
                throw new RuntimeException(e);
            }
        }).map(m -> {
            try {
                return m.getContent().toString();
            } catch (IOException | MessagingException e) {
                throw new RuntimeException(e);
            }
        });
        Parser parser = new Parser();
        Tokenizer tokenizer = new Tokenizer();
        Stream<Stream<Token>> tokens = tokenizer.tokenize(mails);
        Map<Integer, Data> ret = parser.parse(tokens, IntStream.range(1, maxPrefix + 1)
                                                               .boxed()
                                                               .collect(Collectors.toList()));
        System.out.println("generated data map!");
        return ret;
    }
}
