package markov;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.Arrays;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertIterableEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class LookupTests {

    Lookup lookup;

    @BeforeEach
    public void initLookup() {
        lookup = new Lookup();
        lookup.add(new Token("a"));
    }


    @Test
    public void add() {
        lookup.add(new Token("b"));
        assertTrue(lookup.allPossible().contains(new Token("b")));
        assertTrue(lookup.allPossible().contains(new Token("a")));
    }

    @Test
    public void forRandom() {
        assertEquals(lookup.forRandom(0).getToken(), new Token("a"));
        assertEquals(lookup.forRandom(0.5d).getToken(), new Token("a"));
        assertEquals(lookup.forRandom(0.999d).getToken(), new Token("a"));
//        assertEquals(lookup.forRandom(1), new Token("a"));
        assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(1));
        assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(-1));
        assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(1.001d));
        assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(2));

        lookup.resetFinishding();
        lookup.add(new Token("b"));
        Decission a = lookup.forRandom(0);
        Decission b = lookup.forRandom(0.9d);
        assertNotEquals(a, b);

        lookup.resetFinishding();
        lookup.add(new Token("c"));
        a = lookup.forRandom(0);
        b = lookup.forRandom(0.51d);
        Decission c = lookup.forRandom(0.9d);
        assertEquals(new Token("a"), a.getToken());
        assertEquals(new Token("b"), b.getToken());
        assertEquals(new Token("c"), c.getToken());

        lookup.resetFinishding();
        lookup.add(new Token("c"));
        lookup.add(new Token("c"));

        assertEquals(new Token("a"), lookup.forRandom(0).getToken());
        assertEquals(new Token("a"), lookup.forRandom(0.19d).getToken());
        assertEquals(new Token("b"), lookup.forRandom(0.2d).getToken());
        assertEquals(new Token("b"), lookup.forRandom(0.39d).getToken());
        assertEquals(new Token("c"), lookup.forRandom(0.4d).getToken());
        assertEquals(new Token("c"), lookup.forRandom(0.99d).getToken());
    }

    @Test
    public void average() {
        assertEquals(new Token("a"), lookup.average().getToken());
        lookup.resetFinishding();
        lookup.add(new Token("b"));
        Token token = lookup.average().getToken();
        assertTrue(() -> token.equals(new Token("a")) || token.equals(new Token("b")));
        lookup.resetFinishding();
        lookup.add(new Token("a"));
        assertEquals(new Token("a"), lookup.average().getToken());
        lookup.resetFinishding();
        lookup.add(new Token("b"));
        lookup.add(new Token("b"));
        assertEquals(new Token("b"), lookup.average().getToken());

    }

    @Test
    public void allPossible() {
        assertIterableEquals(Arrays.asList(new Token("a")), lookup.allPossible());
        lookup.add(new Token("b"));
        assertIterableEquals(Arrays.asList(new Token("a"), new Token("b")), lookup.allPossible());
        lookup.add(new Token("c"));
        assertIterableEquals(
                Arrays.asList(new Token("a"), new Token("b"), new Token("c")),
                lookup.allPossible());
        lookup.add(new Token("a"));
        assertIterableEquals(
                Arrays.asList(new Token("a"), new Token("b"), new Token("c")),
                lookup.allPossible());
    }

    @Test
    public void getTotalCounts() {
        assertEquals(1, lookup.getTotalCounts());
        lookup.add(new Token("b"));
        assertEquals(2, lookup.getTotalCounts());
        lookup.add(new Token("b"));
        assertEquals(3, lookup.getTotalCounts());
    }

    @Test
    public void getDistinctTokens() {
        assertEquals(1, lookup.getDistinctTokens());
        lookup.add(new Token("b"));
        assertEquals(2, lookup.getDistinctTokens());
        lookup.add(new Token("b"));
        assertEquals(2, lookup.getDistinctTokens());
    }
}