package markov.stuff;

import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.SplittableRandom;
import java.util.stream.Collectors;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertIterableEquals;
import static org.junit.jupiter.api.Assertions.assertNotSame;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

class BitConverterTest {

    private byte[] convert(Boolean... booleans) {
        BitConverter converter = new BitConverter();
        List<Boolean> empty = Arrays.asList(booleans);
        return converter.toBytes(empty);
    }

    private List<Boolean> convert(byte... bytes) {
        BitConverter converter = new BitConverter();
        return converter.toBits(bytes);
    }

    @Nested
    class Both {

        @Disabled
        @Tag("long")
        @Tag("smoke")
        @ParameterizedTest
        @ValueSource(ints = 10_000)
        void firstN(int max) throws IOException, InterruptedException {
            BitConverter converter = new BitConverter();
            SplittableRandom r = new SplittableRandom();
            for (int i = 1; i < max; i++) {
                List<Boolean> list = r.ints(i, 0, 2)
                        .mapToObj(n -> n == 0 ? Boolean.FALSE : Boolean.TRUE)
                        .collect(Collectors.toList());
                List<Boolean> second = converter.toBits(converter.toBytes(converter.toBits(converter
                        .toBytes(list))));
                assertNotSame(list, second);
                assertEquals(list, second);
            }
        }

    }

    @Nested
    class toBytes {

        @Test
        void convertEmpty() {
            Boolean[] bools = new Boolean[0];
            byte[] convertedEmpty = convert(bools);
            assertEquals(1, convertedEmpty.length);
            assertEquals(0b00000011, convertedEmpty[0]);
        }

        @Test
        void convert0() {
            byte[] convertedEmpty = convert(Boolean.FALSE);
            assertEquals(1, convertedEmpty.length);
            assertEquals(0b00000100, convertedEmpty[0]);
        }

        @Test
        void convert1() {
            byte[] convertedEmpty = convert(Boolean.TRUE);
            assertEquals(1, convertedEmpty.length);
            assertEquals(0b00001100, convertedEmpty[0]);
        }

        @Test
        void convert16() {
            byte[] convertedEmpty = convert(
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE);
            assertEquals(1, convertedEmpty.length);
            assertEquals((byte) 0b10000000, convertedEmpty[0]);
        }

        @Test
        void convert31() {
            byte[] convertedEmpty = convert(
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE);
            assertEquals(1, convertedEmpty.length);
            assertEquals((byte) 0b11111000, convertedEmpty[0]);
        }


        @Test
        void convert32() {
            byte[] convertedEmpty = convert(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b00000001, convertedEmpty[0]);
            assertEquals((byte) 0b00000001, convertedEmpty[1]);
        }

        @Test
        void convert63() {
            byte[] convertedEmpty = convert(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b11111001, convertedEmpty[0]);
            assertEquals((byte) 0b00000001, convertedEmpty[1]);
        }

        @Test
        void convert127() {
            byte[] convertedEmpty = convert(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b11111010, convertedEmpty[0]);
            assertEquals((byte) 0b00000011, convertedEmpty[1]);
        }

        @Test
        void convert128() {
            byte[] convertedEmpty = convert(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b00000010, convertedEmpty[0]);
            assertEquals((byte) 0b00000010, convertedEmpty[1]);
        }

        @Test
        void convert4k() {
            byte[] convertedEmpty = convert(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b00000000, convertedEmpty[0]);
            assertEquals((byte) 0b10000000, convertedEmpty[1]);
        }

        @Test
        void convert8k_1() {
            byte[] convertedEmpty = convert(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE);
            assertEquals(2, convertedEmpty.length);
            assertEquals((byte) 0b11111000, convertedEmpty[0]);
            assertEquals((byte) 0b11111111, convertedEmpty[1]);
        }

        @Test
        void convert16k_1() {
            byte[] convertedEmpty = convert(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE);
            assertEquals(3, convertedEmpty.length);
            assertEquals((byte) 0b11111001, convertedEmpty[0]);
            assertEquals((byte) 0b11111111, convertedEmpty[1]);
            assertEquals((byte) 0b00000001, convertedEmpty[2]);
        }
    }

    @Nested
    class toBits {
        @Test
        void convertEmpty() {
            byte[] bytes = new byte[0];
            List<Boolean> converted = convert(bytes);
            assertTrue(converted.isEmpty());
        }

        @Test
        void convertOneByteWIthTooSmallNotEmptyThrows() {
            assertThrows(IllegalArgumentException.class, () -> convert((byte) 0b1));
            assertThrows(IllegalArgumentException.class, () -> convert((byte) 0b10));
            assertThrows(IllegalArgumentException.class, () -> convert((byte) 0b11));
        }

        @Test
        void convertOneByteWithNotEmptyWorks() {
            List<Boolean> result = convert((byte) 0b100);
            assertIterableEquals(Collections.singletonList(Boolean.FALSE), result);

            result = convert((byte) 0b101);
            assertIterableEquals(Arrays.asList(Boolean.FALSE, Boolean.FALSE), result);

            result = convert((byte) 0b110);
            assertIterableEquals(
                    Arrays.asList(Boolean.FALSE, Boolean.FALSE, Boolean.FALSE),
                    result);

            result = convert((byte) 0b111);
            assertIterableEquals(
                    Arrays.asList(Boolean.FALSE, Boolean.FALSE, Boolean.FALSE, Boolean.FALSE),
                    result);

            result = convert((byte) 0b0);
            assertIterableEquals(
                    Arrays.asList(Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.FALSE),
                    result);
        }

        @Test
        void convert0() {
            List<Boolean> result = convert((byte) 0b00000100);
            assertEquals(1, result.size());
            assertIterableEquals(Collections.singletonList(Boolean.FALSE), result);
        }

        @Test
        void convert1() {
            List<Boolean> result = convert((byte) 0b00001100);
            assertEquals(1, result.size());
            assertIterableEquals(Collections.singletonList(Boolean.TRUE), result);
        }

        @Test
        void convert16() {

            List<Boolean> result = convert((byte) 0b10000000);
            assertEquals(5, result.size());
            assertIterableEquals(
                    Arrays.asList(Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.FALSE,
                            Boolean.TRUE),
                    result);

        }

        @Test
        void convert31() {

            List<Boolean> result = convert((byte) 0b11111000);
            assertEquals(5, result.size());
            assertIterableEquals(Arrays.asList(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE), result);
        }


        @Test
        void convert32() {
            List<Boolean> result = convert((byte) 0b00000001, (byte) 0b00000001);
            assertEquals(6, result.size());
            assertIterableEquals(Arrays.asList(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE), result);
        }

        @Test
        void convert63() {
            List<Boolean> result = convert((byte) 0b11111001, (byte) 0b00000001);
            assertEquals(6, result.size());
            assertIterableEquals(Arrays.asList(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE), result);

        }

        @Test
        void convert127() {
            List<Boolean> result = convert((byte) 0b11111010, (byte) 0b00000011);
            assertEquals(7, result.size());
            assertIterableEquals(Arrays.asList(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE), result);

        }

        @Test
        void convert128() {
            List<Boolean> result = convert((byte) 0b00000010, (byte) 0b00000010);
            assertEquals(7, result.size());
            assertIterableEquals(Arrays.asList(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE), result);

        }

        @Test
        void convert4k() {
            List<Boolean> result = convert((byte) 0b00000000, (byte) 0b10000000);
            assertEquals(13, result.size());
            assertIterableEquals(Arrays.asList(Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.FALSE,
                    Boolean.TRUE), result);
        }

        @Test
        void convert8k_1() {
            List<Boolean> result = convert((byte) 0b11111000, (byte) 0b11111111);
            assertEquals(13, result.size());
            assertIterableEquals(Arrays.asList(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE), result);

        }

        @Test
        void convert16k_1() {
            List<Boolean> result = convert((byte) 0b11111001, (byte) 0b11111111, (byte) 0b00000001);
            assertEquals(14, result.size());
            assertIterableEquals(Arrays.asList(Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE,
                    Boolean.TRUE), result);
        }

    }

}